--*******************************************************************************************************************************
--
-- Description: This module initializes all MMCMs and the IDELAY controller in addition to running clock lock on any input pins
--				that require dynamic phase alignment.  The clock lock modules are connected in series through a serial command
--				bus.  Each clock lock process runs sequentially using the clk_lck_start signal to begin and clk_lck_done signal
--				to report completion.
--					
-- Rev History: 09-04-2013:	Inital release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

use work.model_specific_package.all;
use work.transceiver_package.all;

entity clock_init is
port (
	rst 				: in std_logic;		-- Active high reset, logic one places registers in a reset state.
	clk 				: in std_logic;		-- PCI system clock from the Xilinx endpoint core.
	clk200 				: in std_logic;		-- 200 MHz clock from an external oscillator.
	
	clk_init_start 		: in std_logic; 	-- Single period enable to initiate the clock initialization process.
	clk_init_busy		: out std_logic; 	-- Logic one indicates that the clock initialization process is active.
	clk_init_rdy		: out std_logic; 	-- Logic one indicates that the clock initialization completed successfully.
	clk_init_rst		: out std_logic; 	-- Single period active high reset to DCMs and the IODELAY controller.
	clock_status		: in std_logic;		-- Clock status returned from DCMx and the IODELAY controller.
	clk_lck_start 		: out std_logic; 	-- Single period enable to initiate the clock lock process.
	clk_lck_done 		: in std_logic 		-- Single period flag to indicate that the clock lock process is complete.
);
end clock_init;

architecture structural of clock_init is


	type initstate is (idle, reset, settle, dpa, timer, done);--Type definitions for the clock initialization state machine.
	signal init_state		: initstate;					-- Current state of the clock initialization state machine.
	signal last_state		: initstate;					-- Previous state of the clock initialization state machine.
	signal watchdog    		: unsigned(20 downto 0);		-- Watchdog counter to exit the initialization process if hardware does not respond.
	signal dpa_start		: std_logic;					-- Single period enable to intiate the clock lock process (PCI clock domain).
	signal dpa_done			: std_logic; 					-- Single period flag to indicate that the clock lock process is complete (PCI clock domain).
	
begin
	
--************************************************* STATE MACHINE TRANSITIONS **************************************************

	process (clk, rst)
    begin
		if rst='1' then
		
			init_state <= idle;
			
        elsif rising_edge(clk) then
        
			-- Define each of the clock initialization state machine transitions.
			case init_state is
			
				-- Remain idle until an initialization cycle is activated.
				when idle =>
					if clk_init_start = '1' then
						init_state <= reset;			
					end if;
					
				-- Reset all DCMs, the IDELAY controller, and IODELAY elements.
				when reset =>
					init_state <= timer;
					
				-- Wait for the DCMs and IDELAY controller to settle.
				when settle =>					
					init_state <= timer;
				
				-- Run clock lock on any IOBS that require dynamic phase alignment.
				when dpa =>
					init_state <= timer;
					
				-- Wait for the hardware to respond to initialization commands.
				when timer =>
                    -- Hold the reset active (1 us @250 MHz).
					if last_state = reset and watchdog = "000000000000011111111" then
						init_state <= settle;
                    -- Continue to the next state when the DCMs and IDELAY controller lock.
					elsif last_state = settle and clock_status = '1' then
						init_state <= dpa;
                    -- Monitor the clock lock status for the completion of the process.
					elsif last_state = dpa and dpa_done = '1' then
						init_state <= done;
                    -- Transition to the fail state if the watchdog timer expires (8 ms @ 250 MHz).
					elsif watchdog = "111111111111111111111" then
						init_state <= idle;
					end if;
			
				-- Set the ready flag upon successful completion.
				when done =>
					init_state <= idle;
			
				--Define the default state to be idle.		
				when others =>
					init_state <= idle;
				
			end case;	

        end if;
    end process;

--*************************************************** STATE MACHINE EVENTS *****************************************************

	process (clk, rst)
    begin
		if rst='1' then
		
			last_state <= idle;
            watchdog <= (others => '0');
			clk_init_rdy <= '0';
			
        elsif rising_edge(clk) then
        
			-- Save the state preceding the timer state.
			if init_state /= timer then
				last_state <= init_state;
			end if;
			
			-- The watchdog counter is used to exit the timer state if the hardware does not respond.
			if init_state = timer then
				watchdog <= watchdog + "000000000000000000001";
			else
				watchdog <= (others => '0');
			end if;

			-- Generate a ready status flag.
			if init_state = reset then
				clk_init_rdy <= '0';
			elsif init_state = done then
				clk_init_rdy <= '1';
			end if;

        end if;
    end process;

--***************************************************** PORT ASSIGNMENTS *******************************************************

    -- Generate a reset for the DCM and IODELAY elements.
    clk_init_rst <= '1' when last_state = reset else '0';
    -- Generate a state machine busy status flag.
    clk_init_busy <= '0' when init_state = idle else '1';
    
    -- Create a single period start trigger in the 200 MHz clock domain for the daisy-chained clock lock functions.
    dpa_start <= '1' when init_state = dpa else '0';
	LCK_START_CLKXING : entity work.clock_crossing
	port map (
		rst					=> rst,
		source_clk			=> clk,
		target_clk  		=> clk200,
		source_to_target	=> dpa_start,
		target_enable		=> clk_lck_start
	);

    -- Create a single period done trigger in the PCI clock domain from the daisy-chained clock lock functions.
	LCK_DONE_CLKXING : entity work.clock_crossing
	port map (
		rst					=> rst,
		source_clk			=> clk200,
		target_clk  		=> clk,
		source_to_target	=> clk_lck_done,
		target_enable		=> dpa_done
	);

end structural;
